/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef DATAFOX_COMMUNICATION_MODULE_SYSTEM_H
#define DATAFOX_COMMUNICATION_MODULE_SYSTEM_H

// Basisklasse einbinden
#include "datafox/communication/module_hardware.h"
// Verwaltungsklasse fuer BlockTransfer
#include "datafox/communication/controller_segment.h"

#include <boost/ptr_container/ptr_map.hpp>
#include <set>
//#include <sstream>
//#include <string>

#ifdef WITH_VIRTUAL_DEVICE
#include <boost/filesystem.hpp>
#include <fstream>
#include <memory>

#include <boost/format.hpp>
#include <map>
typedef std::map<std::string, std::string> type_variable;
#endif

/** In diesem Module sind alle Systemspezifischen Funktionen enthalten.
 *
 */
#define IV_MODULE_SYSTEM                         0x0001


// Ermitteln der Firmwareversion
#define IV_FUNCTION_GET_VERSION                  0x0000

// Setzen/Ermitteln der Seriennummer
#define IV_FUNCTION_SET_SERIAL                   0x0001
#define IV_FUNCTION_GET_SERIAL                   0x0002
    
// Setzen/Ermitteln der Systemzeit. Format (YYYY-MM-DD HH:MM:SS\0)
#define IV_FUNCTION_SET_RTC                      0x0003
#define IV_FUNCTION_GET_RTC                      0x0004


// Setzen/Ermitteln einer Systemvariablen 
#define IV_FUNCTION_SET_SYSTEM_VARIABLE          0x0005
#define IV_FUNCTION_GET_SYSTEM_VARIABLE          0x0006

// Der Name einer Systemvariablen setzt sich auf folgende Art zusammen:
// MAJOR_NAME.MINOR_NAME
// Für GPRS ist der Majorname MODEM_MC35i.
// Fehlernummer:
// 1 = MAJOR_NAME ist unbekannt
// 2 = MINOR_NAME ist unbekannt
// 3 = Länge des Variablenwertes ist zu lang.


// Setzen/Ermitteln eines Relaiszustandes
#define IV_FUNCTION_SET_RELAY                    0x0007
#define IV_FUNCTION_GET_RELAY                    0x0008

// Die Namen sind: INTERN, TMR (Türmodule), IOM (IOModule)
// Die Busnummer ist die Busnummer bei 0 beginnend. (Bei INTERN eigentlich immer 0)
// Die Relaisnummer ist die Relaisnummer bei 0 beginnend.
// Die Zeit ist eine Angabe in Sekunden (0 bedeutet eine Dauerschaltung in den gewählten Status)
// Fehlernummer:
// 1 = Name undefiniert
// 2 = Busnummer undefiniert
// 3 = Relaisnummer undefiniert
// 4 = Status undefiniert  (0, 1) (Könnte irgendwann um andere Varianten erweitert werden)

// Schreiben/Lesen von: Externen Flash
#define IV_FUNCTION_SET_FLASH                    0x0009
#define IV_FUNCTION_GET_FLASH                    0x000A

// Schreiben/Lesen von: EEPROM
#define IV_FUNCTION_SET_EEPROM                   0x000B
#define IV_FUNCTION_GET_EEPROM                   0x000C

// Schreiben/Lesen von: Program-Flash schreiben/lesen
#define IV_FUNCTION_SET_PROGRAM                  0x000D // (wird nur aufgenommen)
#define IV_FUNCTION_GET_PROGRAM                  0x000E

// Starten/Beenden von: Segmentuebertragung (inklusive aktiver Antwort)
#define IV_FUNCTION_BEGIN_SET_SEGMENT                0x000F
#define IV_FUNCTION_BEGIN_SET_SEGMENT_OK             0x0010
#define IV_FUNCTION_END_SET_SEGMENT                  0x0011
// Im Bestaetigungspaket fuer Befehlerhalt ist der CRC = 0 und 
// im Bestaetigungspaket fuer CRC pruefung ist der CRC = dem bei SYSTEM_START gesendeten.
// Ist der CRC falsch, gibt es ein Fehlerpaket mit Fehlernummer 6.

#ifdef WITH_VIRTUAL_DEVICE
enum class Segment_type
{
    Update,
    Update_background,
    Firmware,
    Fonts,
    Grafics,
    Setup,
    List,
    Setup_extended,
    List_access,
    List_timeboy,
    Fingerprint,
    Text,
    Toch_keys,
    Display_colors,
    Wlan_configs,
    Hardware_configs,
    File_transfer
};

struct Segment_parameter
{
    Segment_type type;
    uint32_t      address;
    uint32_t      size;
    uint32_t      crc;
};

struct Segment_data
{
    Segment_parameter parameter;
    boost::filesystem::path file_path;
    std::shared_ptr<std::fstream> file_stream;
};
#endif

// Die folgenden IV_VALUE_SEGMENT_.. gelten fuer _BEGIN_SEGMENT und _END_SEGMENT
#define IV_VALUE_SEGMENT_UPDATE                            0 // Es werden Daten gesetzt die nachdem Uebernommen werden.
#define IV_VALUE_SEGMENT_UPDATE_BG                         1 // Es werden Daten gesetzt die nachdem im Hintergrund uebernommen werden.
#define IV_VALUE_SEGMENT_FIRMWARE                          2
#define IV_VALUE_SEGMENT_FONTS                             3
#define IV_VALUE_SEGMENT_GRAFICS                           4
#define IV_VALUE_SEGMENT_SETUP                             5
#define IV_VALUE_SEGMENT_LIST                              6
#define IV_VALUE_SEGMENT_SETUP_EXTENDED                    7
#define IV_VALUE_SEGMENT_LIST_ACCESS                       8 // Listen fuer ZK II werden uebertragen, dadurch wird die Zutrittskontrolle blockiert.
#define IV_VALUE_SEGMENT_LIST_TIMEBOY                      9 // Liste fuer Timeboy uebertragen
#define IV_VALUE_SEGMENT_FINGERPRINT                       10
#define IV_VALUE_SEGMENT_TEXT                              11
#define IV_VALUE_SEGMENT_TOUCHKEYS                         12
#define IV_VALUE_SEGMENT_DISPLAYCOLORS                     13
#define IV_VALUE_SEGMENT_WLANCONFIGS                       14
#define IV_VALUE_SEGMENT_HARDWARECONFIGS                   15
#define IV_VALUE_SEGMENT_FILETRANSFER                      16

// 0  -> Global fuer Uebernahmerahmen von Datenbundel mit Bedienstopp. Z. B. Firmware, Setup
// 1  -> Global fuer Uebernahmerahmen von Datenbundel ohne Bedienstopp. Z. B. Listen
// 2  -> Firmware
// 3  -> Fonts
// 4  -> Grafik
// 5  -> Setup
// 6  -> Listen fuer Setup.
// 7  -> Setup Extended Bereich.
// 8  -> Listen fuer Zutrittskontrolle.
// 9  -> Listen fuer Timeboys.
// 10 -> Firmware des Fingerprintmoduls uebertragen
// 11 -> Segment fuer Textdaten der Firmware uebertragen
// 12 -> Segment fuer Daten der Tasten auf Touch übertragen
// 13 -> Segment fuer die Farbdaten des Displays
// 14 -> Segment fuer die WLAN Konfigurationen des Redpine Moduls
// 15 -> Segment für die Hardwarekonfiguration
// 16 -> Segment mit Dateidaten.
// Fehler:
// 1 -> Unbekannter SS-Type
// 2 -> Zu viele Updatesegmente (7 vordefiniert)
// 3 -> Angeforderter Speicherplatz steht nicht zur Verfügung
// 4 -> Kein Stopp vor Start
// 5 -> Verschachtelungsfehler von Stopp und Start. {[]()} -> OK, aber {[( -> Fehler
// 6 -> CRC Berechnung schlug fehl.

#define IV_FUNCTION_END_SET_SEGMENT_OK               0x0012
// Wird vom Geraet aufgrund eines SYSTEM_START -> SYSTEM_END mit nachfolgender Berechnung vom CRC gesendet.

// Ermitteln von: Bootloaderversion
#define IV_FUNCTION_GET_VERSION_BOOTLOADER       0x0013
// Format: XX.XX.XX??????????????

// Durchfuehren eines Resets der Speicherbereiche
#define IV_FUNCTION_SYSTEM_RESET                 0x0014

// Resetmodus:
#define IV_VALUE_RESET_ALL             0 // Geraet in einen Zustand versetzen, wie nach dem Bootloader einspielen.
#define IV_VALUE_RESET_RESTART         1 // Warmstart (Watchdog ausführen)
#define IV_VALUE_RESET_FORMAT          2 // Flash formatieren.
#define IV_VALUE_RESET_SETUP           3 // Setup, Listen und Datensatzzeiger Löschen/Zurücksetzen

// Fehler:
// 1 -> Resetmodus unbekannt.


// Ins Bios wechseln
#define IV_FUNCTION_ENTER_BIOS			         0x0015
// Bioswechsel erfolgt
#define IV_FUNCTION_ENTERED_BIOS		         0x0016

// Bios verlassen (aenderungen uebernehmen
#define IV_FUNCTION_EXIT_BIOS			         0x0017
// Biosverlassen erfolgt
#define IV_FUNCTION_EXITED_BIOS			         0x0018

// Setzt/liest das zu verwendende Interface
#define IV_FUNCTION_SET_INTERFACE                0x0019
// 0 -> RS232
// 1 -> TCP/IP
// 2 -> GSM
// 3 -> GPRS
// 4 -> RS485

// 9 -> USB-Host
#define IV_FUNCTION_GET_INTERFACE                0x001A

#define IV_VALUE_INTERFACE_RS232       0
#define IV_VALUE_INTERFACE_TCPIP       1
#define IV_VALUE_INTERFACE_GSM         2
#define IV_VALUE_INTERFACE_GPRS        3
#define IV_VALUE_INTERFACE_RS485       4
#define IV_VALUE_INTERFACE_USB_HOST    9


// Setzen/Ermitteln der zu verwendende BaudRate
#define IV_FUNCTION_SET_BAUD                     0x001B
#define IV_FUNCTION_GET_BAUD                     0x001C
// Setzen/Ermitteln der zu verwendenden DeviceID
#define IV_FUNCTION_SET_DEVICE_ID                0x001D
#define IV_FUNCTION_GET_DEVICE_ID                0x001E

// Befehl der beim Geraetestart den Bootloader des MasterIV wiederherstellt.
#define IV_FUNCTION_ENTER_BOOTLOADER             0x001F

// In Monitor Mode von XPort wechselt -> siehe auch #define IV_FUNCTION_ENTER_XPORT_MONITOR_EXT		 0x0028 
#define IV_FUNCTION_ENTER_XPORT_MONITOR          0x0020
// Antwort in Monitor Mode gewechselt -> siehe auch #define IV_FUNCTION_ENTER_XPORT_MONITOR_EXT_OK   0x0029
#define IV_FUNCTION_ENTER_XPORT_MONITOR_OK       0x0021

// Monitorbefehl an XPort und Antwort erhalten.
#define IV_FUNCTION_CMD_XPORT                    0x0022

// In Monitor Mode von XPort wechselt -> siehe auch #define 
#define IV_FUNCTION_EXIT_XPORT_MONITOR           0x0023
// Antwort in Monitor Mode gewechselt -> siehe auch #define 
#define IV_FUNCTION_EXIT_XPORT_MONITOR_OK        0x0024


#define IV_FUNCTION_BEGIN_GET_SEGMENT            0x0025
// [in]-> [UINT16(SS_...)]
// <- 16bit 
#define IV_FUNCTION_BEGIN_GET_SEGMENT_OK         0x0026
// [out]-> 16bit Typ des Speichers
// [out]-> 32bit Adresse fuer Lesen
// [out]-> 32bit Groesse des Blocks
// [out]-> 32bit CRC des Blocks

#define IV_FUNCTION_END_GET_SEGMENT              0x0027
// -> [UINT16(SS_...)]
// <- [UINT16(SS_...)]
// Im Bestaetigungspaket fuer Befehlerhalt ist der CRC = 0 und 
// im Bestaetigungspaket fuer CRC pruefung ist der CRC = dem bei SYSTEM_START gesendeten.
// Ist der CRC falsch, gibt es ein Fehlerpaket mit Fehlernummer 6.

// Die folgenden IV_VALUE_SEGMENT_.. gelten fuer _BEGIN_GET_SEGMENT und _END_GET_SEGMENT
#define IV_VALUE_SEGMENT_GET_UPDATE         0 // Es werden Daten gelesen ohne Bedienstop.
#define IV_VALUE_SEGMENT_GET_UPDATE_BG      1 // Es werden Daten gesetzt die nachdem im Hintergrund uebernommen werden.
#define IV_VALUE_SEGMENT_GET_SETUP          5
#define IV_VALUE_SEGMENT_GET_SETUP_EXTENDED 7

// Fehler:
// 1 -> Unbekannter SS-Type
// 2 -> Zu viele Updatesegmente (7 vordefiniert)
// 3 -> Angeforderter Speicherplatz steht nicht zur Verfügung
// 4 -> Kein Stopp vor Start
// 5 -> Verschachtelungsfehler von Stopp und Start. {[]()} -> OK, aber {[( -> Fehler
// 6 -> CRC Berechnung schlug fehl.

#define IV_FUNCTION_END_GET_SEGMENT_OK           0x0028
// Wird vom Geraet aufgrund eines SYSTEM_START -> SYSTEM_END mit nachfolgender Berechnung vom CRC gesendet.

#define IV_FUNCTION_SYSTEM_RESET_OK              0x0029

// Transparentmodus aktivieren/deaktivieren.
#define IV_FUNCTION_OPEN_COM_SERVER_MODE         0x002A
#define IV_FUNCTION_CLOSE_COM_SERVER_MODE        0x002B

// Abfrage von Information zu einem Modultyp
#define IV_FUNCTION_GET_INFO                     0x002C
// param1: ACCESS.MODULE
// param2: bei ACCESS.MODULE UCHAR(BusNr) UCHAR(0=TM,1=L)

// Antwort in Com-Server-Mode gewechselt
#define IV_FUNCTION_OPEN_COM_SERVER_MODE_OK      0x002D

// Aktivieren/Deaktivieren und Abfragen der Benachrichtigungen
#define IV_FUNCTION_SET_NOTIFY                   0x002E
// param1: Typ der Benachrichtigung
// param2: Zu setztender Zustand der Benachrichtigung
#define IV_FUNCTION_GET_NOTIFY                   0x002F
// param1: Typ der Benachrichtigung
// Werte siehe IV_MODULE_APPLICATION.IV_FUNCTION_NOTIFY

#define IV_FUNCTION_GET_SOFTWARE_INFO            0x0030

// Lesen einer Seite des Flash (Bereich: 0 - 4095, Seite hat 528Byte)
#define IV_FUNCTION_GET_FLASH_PAGE               0x0031

// Setzen/Lesen der Systemflags von EEPROM-Adresse: 0x0ECC
// Bit1: Vorhandenes Setup und Listen werden übernommen und durch das Update nicht gelöscht.
//       Nur Mobil-MasterIV. Die Mindestversion auf dem Gerät muß 04.01.04.32 sein und die aufzuspielende mindestens 04.01.04.59 damit diese Option korrekt unterstützt wird.;
// Bit2-32: Reserved
#define IV_FUNCTION_SET_SYSTEM_FLAGS             0x0032
#define IV_FUNCTION_GET_SYSTEM_FLAGS             0x0033

#define IV_VALUE_SYSTEM_FLAG_MASK                0x0000FFFF
#define IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS 0x00000001


// Schreiben/Lesen von: RAM
#define IV_FUNCTION_SET_RAM                      0x0034
#define IV_FUNCTION_GET_RAM                      0x0035

// Schreiben/Lesen von MMC
#define IV_FUNCTION_SET_MMC                      0x0036
#define IV_FUNCTION_GET_MMC                      0x0037
#define IV_FUNCTION_GET_SFSMPB                   0x0038

#pragma  pack(1)
// Strucktur fuer dynamische Verwaltung des SFS (MS) 06.11.08
typedef struct struct_memory_partition_block
{
    uint32_t  StorageStartAddress; ///< Startadresse des Filessystems im Speicher
    uint32_t  StartFATAddress;     ///< Startadresse FAT
    uint32_t  DirStartAddress;     ///< Startadresse File Directory
    uint32_t  FileStartAddress;    ///< Startadresse Datenbereich
    uint32_t  FirstRecordAddress;  ///< Startadresse des Ringspeichers
    uint32_t  StorageEndAddress;   ///< Letzte Adresse des Speicherbereiches

    uint16_t  DirStartCluster;     ///< Start Cluster in dem das Dateiverzeichnis liegt
    uint16_t  FileStartCluster;    ///< Start Cluster für die Daten der Dateien
    uint16_t  FATClusterCount;     ///< Anzahl der Cluster die von der FAT belegt werden
    uint16_t  DirClusterCount;     ///< Anzahl der Cluster die vom Verzeichnis belegt werden.
    uint16_t  FileClusterCount;    ///< Anzahl der Cluster für den Dateibereich
    uint16_t  TotalClusterCount;   ///< Anzahl aller Cluster des Dateisystems
    uint16_t  SectorsPerCluster;   ///< Anzahl der Sektoren pro Cluster

    uint32_t  ClusterSize;         ///< Größe eines Clusters in Byte
    uint32_t  RecordVolumeSize;    ///< Größe des Ringspeichers in Byte
    uint32_t  ListVolumeSize;      ///< Größe des Dateisystem in Byte
    uint32_t  StorageVolumeSize;   ///< Größe der Memory-Card in Byte

    uint8_t ClusterShiftBit;     ///< Schiebebit für Clusteradressierung

} SFSMPB;
#pragma pack()

// MatchPort in den Monitor Mode (z-Mode mit aktivem WLAN Interface) wechseln
#define IV_FUNCTION_ENTER_XPORT_MONITOR_EXT		 0x0039
// Antwort auf den Wechsel-Befehl BetriebsMode->MonitorMode
#define IV_FUNCTION_ENTER_XPORT_MONITOR_EXT_OK   0x003A

// System Device Log
#define IV_FUNCTION_GET_SYS_RECORD				 0x003B
#define IV_FUNCTION_SET_SYS_RECORD				 0x003C
#define IV_FUNCTION_RESTORE_ALLSYS_RECORDS		 0x003D

#define IV_ANSW_NO_SYS_RECORD               0xC0
#define IV_ANSW_ONE_SYS_RECORD				0xC1
#define IV_ANSW_ACK_ON_SYS_RECORD			0xC8
#define IV_SYS_RECORDS_MAX					0x200

// Passwort fuer Kommunikation
#define IV_FUNCTION_PASSWORD_CHECK              0x003E
#define IV_FUNCTION_PASSWORD_KEY                0x003F

#define IV_FUNCTION_PRESS_VIRTUAL_KEY           0x0040

// Schreiben eines Datenstroms
#define IV_FUNCTION_TRANSMIT_STREAM             0x0041
#define IV_FUNCTION_TRANSMIT_STREAM_OK          0x0042

#define IV_FUNCTION_RECEIVE_STREAM              0x0043

enum class Stream_type
{
    Update = 0,
    Update_background,
    Firmware,
    Fonts,
    Graphics,
    Setup,
    Setup_list,
    Setup_extended,
    List_access,
    List_timeboy,
    Fingerprint,
    Text,
    Touch_keys,
    Display_colors,
    Wlan_configs,
    Fake
};

// Passwort fuer Systemkonfigurationen
#define IV_FUNCTION_SYSTEM_KEY_CHECK          0x0044
#define IV_FUNCTION_SYSTEM_KEY                0x0045

// Lesen der Flashdaten über die physikalische Adresse
#define IV_FUNCTION_GET_FLASH_PHYSICAL        0x0046

// Befehl zum Transparenten durchsenden von Daten.
// Hierbei werden die MasterIV-spezifischen Zeichen wie '~' maskiert gesendet,
// der Empfänger demaskiert diese je nach Param2-Flag, ebenso bei dem Empfang.
#define IV_FUNCTION_TRANSPARENCY_MODE         0x0047
// Param1: uint16_t Slot-Nummer oder id zu der durchgeschaltet werden soll.
// 1 - 254 -> Slotnummer
// 255 - .. -> Andere Bedeutung. Daten zur Id aus der bios_settings.
// Param2: uint16_t Flags
// Bit1: Wenn gesetzt (Raw-Modus) werden alle Zeichen 1:1 durchgeleitet. Ansonsten (Masked-Modus) werden die empfangenen bzw. zu sendenden Zeichen de-/maskiert.
// Bit2: Wenn gesetzt wird die Zeitangabe in Param5 in Sekunden gewertet, ansonsten in Millisekunden.
// Bit3: Wenn gesetzt sollen nur Timeboy I-III Geräte angesprochen werden. Da TimeboyIV Geräte einen TimeboyIII in ihrer Kommunikation nachahmen können
//       wird dieses Flag dazu verwendet, Bereits beim Herstellen des TransparentMode ältere Timeboystände zu verlangen. Nutzung im TbSetup und TbCom.dll
// Bit4-16: Unbenutzt
// Param3: uint32_t gewünschte Lesepuffergröße der Gegenstelle in Byte.
// Param4: uint32_t gewünschte Schreibpuffergröße der Gegenstelle in Byte.
// Param5: uint16_t Timeout bei Kommunikation über Raw-Modus in Milisekunden oder Sekunden, je nach Flag-Bit2.

// Befehle zum Schalten der gelben LED bei Dockingstation-V2
#define IV_FUNCTION_SET_LED                   0x0048
#define IV_FUNCTION_GET_LED                   0x0049
// Param1: uint8_t slot. Welcher Slot angesprochen werden soll.
// 0 -> Das Gerät mit der deviceId selbst.
// 1 - 253 -> Slotnummer
// Param2: uint8_t Led-Id welche LED angesprochen wird.
// Bei der Mehrfachdocking entspricht diese Id folgenden LED's.
// 1 -> Rote
// 2 -> Gelbe
// 3 -> Grüne
// Param3: uint8_t Status der Blinkfrequenz.
// Hierbei handelt es sich um die Prozentangabe der Zeit relativ zu einer Sekunde.
// 6,66% sind 66,6ms an, 933,4ms aus.
// 93,33% sind 933,3ms an, 66,7ms aus.
#define IV_VALUE_LED_OFF             ( 0x00 )    // Flag LedOff
#define IV_VALUE_LED_PERCENT01       ( 0x01 )    // Flag LedOn6,66%
#define IV_VALUE_LED_PERCENT02       ( 0x02 )    // Flag LedOn13,33%
#define IV_VALUE_LED_PERCENT03       ( 0x03 )    // Flag LedOn19,99%
#define IV_VALUE_LED_PERCENT04       ( 0x04 )    // Flag LedOn26,66%
#define IV_VALUE_LED_PERCENT05       ( 0x05 )    // Flag LedOn33,33%
#define IV_VALUE_LED_PERCENT06       ( 0x06 )    // Flag LedOn39,99%
#define IV_VALUE_LED_PERCENT07       ( 0x07 )    // Flag LedOn46,66%
#define IV_VALUE_LED_PERCENT08       ( 0x08 )    // Flag LedOn53,33%
#define IV_VALUE_LED_PERCENT09       ( 0x09 )    // Flag LedOn59,99%
#define IV_VALUE_LED_PERCENT10       ( 0x0A )    // Flag LedOn66,66%
#define IV_VALUE_LED_PERCENT11       ( 0x0B )    // Flag LedOn73,33%
#define IV_VALUE_LED_PERCENT12       ( 0x0C )    // Flag LedOn79,99%
#define IV_VALUE_LED_PERCENT13       ( 0x0D )    // Flag LedOn86,66%
#define IV_VALUE_LED_PERCENT14       ( 0x0E )    // Flag LedOn93,33%
#define IV_VALUE_LED_ON              ( 0x0F )    // Flag LedOn
// Bei setLED wird Param1 + Param2 + Param3 gesendet und als Antwort keiner erwartet.
// Bei getLED wird Param1 + Param2 gesendet und als Antwort Param1 + Param2 + Param3 erwartet.

// Active-Answer bei Abfrage von Information zu einem Modultyp
// Wenn in der Antwort auf GET_INFO als Information "Active-Answer" (case insensitiv) gesendet wird,
// dann wird im folgenden mittels Aktiver-Antwort auf die Information gewartet.
#define IV_FUNCTION_GET_INFO_OK                  0x004A

namespace Datafox
{
    namespace Communication
    {
        class Module_system : public Module_hardware
        {
        public:
            Module_system( int connection_id = Connection_base::dv_connection_id, uint8_t device_address = Protocol_base::dv_device_address );
            virtual ~Module_system();

        public:
            inline void set_password_communication( const std::string &password )
            {
                m_password_communication = password;
            }
            inline const std::string &get_password_communication()
            {
                return m_password_communication;
            }
            inline void set_flag_password_protected( int state )
            {
                m_is_password_protected = (state == 0) ? 0 : 1;
            }
            // Ist Passwortschutz aktiv?
            inline int is_password_protected()
            {
                return m_is_password_protected;
            }
            inline void set_flag_password_protection( int state )
            {
                m_is_password_protection = (state == 0) ? 0 : 1;
            }
            // Wird Passwortschutz unterstützt?
            inline int is_password_protection()
            {
                return m_is_password_protection;
            }

        private:
            std::string m_password_communication; // Passwort fuer die Kommunikation
            int m_is_password_protected; // Haelt die Information, dass das Geraet Passwortgeschuetzt ist.
            int m_is_password_protection; // Haelt die Information, dass das Geraet den Passwortschutz unterstuetzt.

        public:
            std::string m_version_firmware;

        public:
            int write_data( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials = Connection_base::dv_write_trials );
            int read_data( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials = Connection_base::dv_read_trials );
            enum enum_trace_data_type
            {
                tdt_write,
                tdt_read
            };
            enum enum_trace_data_mode
            {
                tdm_full,
                tdm_full_with_crlf,
                tdm_begin,
                tdm_append,
                tdm_end,
                tdm_end_with_crlf
            };
            void trace_data( uint8_t *data, std::size_t size, enum_trace_data_mode mode = tdm_full_with_crlf, const char *prefix = nullptr );

        public:
            // Gibt Modulfehler zurueck.
            virtual void error_module( uint16_t module, uint16_t error, std::string &text );
            // Gibt Funktionsfehlertext zurueck.
            virtual void error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text );

            // MODULE_HARDWARE
            // BUG:
            // Im Bootloader 00.04.00.00 && 00.04.00.01 liefert die Funktion
            // - TypeTransponder den Wert fuer das Display
            // - TypeDisplay den Wert fuer das Keypad
            // - TypeKeypad den Wert fuer den Transponder
            int write_display_type( uint8_t device_address, uint8_t val, uint16_t value2 = 0xFFFF, bool checked = false );
            int read_display_type( uint8_t device_address, uint8_t &val, bool checked = false );
            int write_keypad_type( uint8_t device_address, uint8_t val, uint16_t value2 = 0xFFFF, bool checked = false );
            int read_keypad_type( uint8_t device_address, uint8_t &val, bool checked = false );
            int write_transponder_reader_type( uint8_t device_address, uint8_t val, uint16_t value2 = 0xFFFF, bool checked = false );
            int read_transponder_reader_type( uint8_t device_address, uint8_t &val, bool checked = false );

            // Gesamte Hardwareinfo lesen
            int read_hardware_info( uint8_t device_address, StructHardwareInfo &hardware_info );
            // Gesamte Hardwareinfo in String umsetzen.
            int hardware_info_to_string( StructHardwareInfo hardware_info, std::string &info );
    
            // MODULE_SYSTEM

            // Systemreset; Speicherbereiche loeschen
            int system_reset( uint8_t device_address, uint16_t modus );
            int reply_system_reset( uint8_t device_address );

            // Firmwareversion lesen
            int read_firmware_version( uint8_t device_address, std::string &version );
            int read_bootloader_version( uint8_t device_address, std::string &version );

            // Seriennummer lesen
            int write_serial_number( uint8_t device_address, uint32_t serial, uint32_t value2 = 0xFFFFFFFF );
            virtual int read_serial_number( uint8_t device_address, uint32_t &serial );

            // System Device Log
            int read_system_record( uint8_t device_address, uint8_t *buffer, std::size_t &count, uint8_t &status );
            int quit_system_record( uint8_t device_address, uint8_t &status );
            int restore_system_records( uint8_t device_address );

            // Datum/Uhrzeit setzen/lesen
            int write_date_time( uint8_t device_address, char* time );
            int read_date_time( uint8_t device_address, char* time );

            // Helper für Systemvariablen von Mobile
            void latin1_to_gsm_codec(std::string &buf );
            void gsm_to_latin1_codec(std::string &buf);
            uint8_t hex_to_uint8( char* data );

            // Systemvariable setzen/lesen
            int write_system_variable( uint8_t device_address, const std::string &name, const std::string &val );
            int read_system_variable( uint8_t device_address, const std::string &name, std::string &val );

            // Relais setzen/lesen
            int write_relay_state( uint8_t device_address, const std::string &name, uint8_t bus_number, uint8_t relay_number, uint8_t status, uint16_t time );
            int read_relay_state( uint8_t device_address, const std::string &name, uint8_t bus_number, uint8_t relay_number, uint8_t &status, uint16_t &time );

        protected:
            uint32_t m_crc_write_segment;
            uint32_t m_crc_read_segment;
            std::string m_stream_buffer;
            bool m_stream_transfer;
        public:
            void set_stream_transfer( bool state )
            { 
                m_stream_transfer = state;
            }
            bool is_stream_transfer()
            { 
                return m_stream_transfer;
            }
            void clear_stream_transfer()
            { 
                m_stream_buffer.clear();
            }

        protected:
            std::mutex m_mutex_segment;      // Segmentuebertragung

        public:
            std::set<uint8_t> get_controller_segment_keys()
            {
                std::set<uint8_t> keys;
                for ( auto iter = m_map_controller_segment.cbegin(); iter != m_map_controller_segment.cend(); ++iter )
                {
                    keys.insert(iter->first);
                }

                return keys;
            }

            inline int get_controller_segment_active_count( controller_segment::block_transfer_type blockType = controller_segment::btt_mask_known )
            {
                int counter = 0;
                std::set<uint8_t> keys( get_controller_segment_keys() );
                for ( auto iter = keys.cbegin(); iter != keys.cend(); ++iter )
                {
                    // Globale-Object uebergehen.
                    if ( *iter == 255 )
                    {
                        continue;
                    }

                    if ( (blockType == controller_segment::btt_mask_known || blockType == m_map_controller_segment[*iter].transfer_state) 
                        && m_map_controller_segment[*iter].transfer_state == controller_segment::bts_active )
                    {
                        ++counter;
                    }
                }

                return counter;
            }

            // Setzt in allen ControllSegment die ggf. vom Global-Objekt uebernommene duration auf 0.
            inline void reset_controller_segment_global_duration()
            {
                std::set<uint8_t> keys( get_controller_segment_keys() );
                for ( auto iter = keys.cbegin(); iter != keys.cend(); ++iter )
                {
                    // Globale-Object uebergehen.
                    if ( *iter == 255 )
                    {
                        continue;
                    }

                    if ( m_map_controller_segment[*iter].transfer_flags & controller_segment::cf_settings_from_global )
                    {
                        m_map_controller_segment[*iter].allowed_transfer_types = controller_segment::btt_none;
                        m_map_controller_segment[*iter].duration = std::chrono::seconds(0);
                        m_map_controller_segment[*iter].transfer_flags &= ~controller_segment::cf_settings_from_global;
                    }
                }
            }

            inline controller_segment &get_controller_segment( uint8_t device_address )
            {
                return m_map_controller_segment[device_address];
            }

        protected:
            boost::ptr_map<uint8_t, controller_segment> m_map_controller_segment;

        public:
            // Datensegmentuebertragung beginnen/enden
            int enter_write_data_segment( uint8_t device_address, uint16_t art, uint32_t &address, uint32_t size, uint32_t crc );
            int reply_enter_write_data_segment( uint8_t device_address );
            int exit_write_data_segment( uint8_t device_address, uint16_t art, uint32_t crc );
            int reply_exit_write_data_segment( uint8_t device_address );

            // Datensegmentuebertragung beginnen/enden
            int enter_read_data_segment( uint8_t device_address, uint16_t art, uint32_t &address, uint32_t &size, uint32_t &crc );
            int reply_enter_read_data_segment( uint8_t device_address );
            int exit_read_data_segment( uint8_t device_address, uint16_t art );
            int reply_exit_read_data_segment( uint8_t device_address );

            // Flash setzen/lesen mittels logischer Adresse (die ersten 16kByte Bootloader koennen nur gelesen werden)
            int write_to_flash( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_from_flash( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_flash_page( uint8_t device_address, std::size_t page_number, std::size_t offset, uint8_t *buffer, std::size_t &count );
            // Flashdaten lesen, mittels physikalischer Adresse.
            int read_from_flash_physical( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );

            // EEPROM setzen/lesen
            int write_to_eeprom( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_from_eeprom( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );

            // PROGRAM setzen/(lesen unterbunden)
            int write_to_program_flash( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_from_program_flash( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );

            // RAM setzen/lesen
            int write_to_ram( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_from_ram( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );

            // MMC schreiben/lesen
            int write_to_mmc( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );
            int read_from_mmc( uint8_t device_address, std::size_t address, uint8_t *buffer, std::size_t count );

            // Info zum Dateisystem
            int read_sfs_mpb( uint8_t device_address, uint8_t &type, SFSMPB &sfs );

            // Systembios aufrufen/verlassen
            int enter_bios_mode( uint8_t device_address );
            int reply_enter_bios_mode( uint8_t device_address );
            int exit_bios_mode( uint8_t device_address );
            int reply_exit_bios_mode( uint8_t device_address );

            // Interface setzen/lesen (RS232, TCPIP...)
            int write_main_interface_type( uint8_t device_address, uint8_t type );
            int read_main_interface_type( uint8_t device_address, uint8_t &type );
    
            // Baud-Rate setzen/lesen
            int write_baud( uint8_t device_address, uint32_t baud );
            int read_baud( uint8_t device_address, uint32_t &baud );

            // DeviceID setzen/lesen
            int write_device_address( uint8_t device_address, uint16_t address );
            int read_device_address( uint8_t device_address, uint16_t &address );

            // Bootloader aktivieren
            int enter_bootloader_mode( uint8_t device_address);

            // Monitor Mode vom XPort
            int enter_xport_monitor_mode( uint8_t device_address, char* data, uint16_t &count );
            int reply_enter_xport_monitor_mode( uint8_t device_address );
            int enter_xport_monitor_mode_ext( uint8_t device_address, char* data, uint16_t &count );
            int reply_enter_xport_monitor_mode_ext( uint8_t device_address );
            int xport_monitor_mode_command( uint8_t device_address, char* buffer, uint16_t &count );
            int exit_xport_monitor_mode( uint8_t device_address );
            int reply_exit_xport_monitor_mode( uint8_t device_address );

            // Transparentsmodus fuer COM
            int enter_com_server_mode( uint8_t device_address, const char* comm, uint32_t baud, uint8_t byte_size, uint8_t parity, uint8_t stop_bits, uint16_t flags, uint16_t timeout_to_close );
            int reply_enter_com_server_mode( uint8_t device_address );
            int exit_com_server_mode( uint8_t device_address );

            // Informationen zu einem Modultyp abfragen
            int read_hardware_module_info(uint8_t device_address, const std::string &name, const std::string &param, std::string &info);
            int reply_read_module_info( uint8_t device_address );

            // Aktivieren/Deaktivieren und Abfragen der Benachrichtigungen
            int write_notify( uint8_t device_address, uint8_t type, uint8_t state );
            int read_notify( uint8_t device_address, uint8_t type, uint8_t &state );

            // Softwareinformation der Anwendungsfirmware abfragen
            int read_software_info( uint8_t device_address, uint8_t* info, int &info_length );

            // Setzen/Lesen der Systemflags
            int write_system_flags( uint8_t device_address, uint32_t flags );
            int read_system_flags( uint8_t device_address, uint32_t &flags );
    
            // Passwort pruefen
            int write_setup_password( uint8_t device_address, const std::string &password );
            // Passwortschlüssel abrufen
            int read_setup_password_key( uint8_t device_address, char* key );

            // Einen Tastencode ausführen. F1-F15 drücken.
            int write_virtual_key( uint8_t device_address, uint8_t key_code, uint8_t flags );

            // Daten als ein Datenstrom senden.
            int write_stream( uint8_t device_address, Stream_type type, const char *buffer, std::size_t count );
            // Daten als ein Datenstrom empfangen.
            int read_stream( uint8_t device_address, Stream_type type, char *buffer, std::size_t count );

            // Systemschlüssel pruefen
            int write_system_key( uint8_t device_address, char* key );
            // Systemschlüssel abrufen
            int read_system_key( uint8_t device_address, char* key );

        private:
            enum enum_tranparency_mode
            {
                tm_off,    // Nicht aktiv.
                tm_masked, // Aktiv, wobei Sonderzeichen maskiert werden.
                tm_raw     // Aktiv, alles wird 1:1 weitergeleitet.
            };
            enum_tranparency_mode m_transparency_mode;
        public:
            // Daten transparent übertragen.
            int enter_transparency_mode( uint8_t device_address, uint16_t id, uint16_t flags, uint32_t read_buffer_size, uint32_t write_buffer_size, int timeout );
            int exit_transparency_mode( uint8_t device_address );
            bool is_transparency_mode();

            // Gelbe LED bei Dockingstation-V2 setzen/lesen
            int write_led_state( uint8_t device_address, uint8_t slot, uint8_t id, uint8_t state);
            int read_led_state( uint8_t device_address, uint8_t slot, uint8_t id, uint8_t &state);

        public:
            // Ausfuehren eines Protokollbefehls
            virtual int call( uint16_t context, uint16_t module, uint16_t function, uint8_t *buffer, uint16_t &count, int context_value = -1 );

            //#################################
            // remoteCall - Routinen

            // Ausfuehren von Befehlen
            virtual int remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count );

        #ifdef WITH_VIRTUAL_DEVICE
        protected:
            type_variable m_variables;
            std::vector<Segment_data> m_segments;
        #endif
        };

    } // namespace Communication
} // namespace Datafox

#endif
