/**
* @file  file_wlan_config.cpp
* @brief Implementierung der Klasse FileWlanConfig.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_wlan_config.h"

#include <string>

FileWlanConfig::FileWlanConfig()
{
    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::wlan_config );

    // Initalisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();
}

FileWlanConfig::~FileWlanConfig()
{
}

void FileWlanConfig::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Pruefsumme zum Test auf Datenaenderung ruecksetzen.
    m_crc = 0;

    // Member initalisieren
    m_wlan_config_node = nullptr;
}

void FileWlanConfig::init()
{
    // Speicher fuer die Daten.
    m_buffer.resize( sizeof( WLANCONFIGNODE ) + max_wlan_configs * sizeof( unsigned int ) + max_wlan_configs * sizeof( WLANCONFIG ) );

    // Hilfsstrukturzeiger
    m_wlan_config_node = (WLANCONFIGNODE *)&m_buffer[0];

    // 1.
    // WlanConfig init
    // Grundstruktur aufsetzen
    m_wlan_config_node->version = 1;
    m_wlan_config_node->cnt = max_wlan_configs;
    m_wlan_config_node->crypt = 0;
    m_wlan_config_node->defaultConfig = 0;
    m_wlan_config_node->flags = 0x01; // Automatisch die Konfiguration waehlen.
    m_wlan_config_node->ptrWlanConfigArray = sizeof( WLANCONFIGNODE );


    // Zeigerarray und WlanConfigstrukturzeiger initialisieren
    std::size_t i = m_wlan_config_node->cnt;
    for ( std::size_t j = 0; j < i; ++j )
    {
        // Strukturoffset auf WLAN Konfiguration
        *(unsigned int *)&m_buffer[m_wlan_config_node->ptrWlanConfigArray + j * sizeof( unsigned int )] =
            m_wlan_config_node->ptrWlanConfigArray + i * sizeof( unsigned int ) + j * sizeof( WLANCONFIG );

        m_wlan_config[j] = (WLANCONFIG *)&m_buffer[m_wlan_config_node->ptrWlanConfigArray + i * sizeof( unsigned int ) + j * sizeof( WLANCONFIG )];

        DefaultWlanConfig( *m_wlan_config[j] );
    }
}

void FileWlanConfig::SetWlanConfig( int idx, const WLANCONFIG &wc )
{
    *m_wlan_config[idx] = wc;
}

WLANCONFIG FileWlanConfig::GetWlanConfig( int idx )
{
    return *m_wlan_config[idx];
}

void FileWlanConfig::WlanConfigCRC( int idx )
{
    // Alle Verfuegbaren.
    if ( idx == -1 )
    {
        idx = GetWlanConfigCount();
        for ( int i = 0; i < idx; ++i )
        {
            m_wlan_config[i]->crc = crc16( (unsigned char *)m_wlan_config[i], sizeof( WLANCONFIG ) - sizeof( unsigned short ) );
        }
    }
    else if ( idx >= 0 && idx < GetWlanConfigCount() )
    {
        m_wlan_config[idx]->crc = crc16( (unsigned char *)m_wlan_config[idx], sizeof( WLANCONFIG ) - sizeof( unsigned short ) );
    }
}

bool FileWlanConfig::WlanConfigCheckCRC( int idx )
{
    bool res = true;
    // Alle Verfuegbaren pruefen.
    if ( idx == -1 )
    { 
        idx = GetWlanConfigCount();
        for ( int i = 0; i < idx && res; ++i )
        {
            res = m_wlan_config[i]->crc == crc16( (unsigned char *)m_wlan_config[i], sizeof( WLANCONFIG ) - sizeof( unsigned short ) );
        }
    }
    else
    {
        res = m_wlan_config[idx]->crc == crc16( (unsigned char *)m_wlan_config[idx], sizeof( WLANCONFIG ) - sizeof( unsigned short ) );
    }

    return res;
}

void FileWlanConfig::WlanConfigCrypt( int idx )
{
    // !!! ACHTUNG: Routine verwendet zuvor zu setzenden Schluessel !!!

    // Alle Verfuegbaren.
    if ( idx == -1 )
    { 
        idx = GetWlanConfigCount();
        for ( int i = 0; i < idx; ++i )
        {
            m_crypt.init();
            m_crypt.crypt( (unsigned char *)m_wlan_config[i], sizeof( WLANCONFIG ) );
        }
    }
    else if ( idx >= 0 && idx < GetWlanConfigCount() )
    {
        m_crypt.init();
        m_crypt.crypt( (unsigned char *)m_wlan_config[idx], sizeof( WLANCONFIG ) );
    }
}

bool FileWlanConfig::load( const std::string &filePath )
{
    // Lesen der Hexdatei
    if ( FileIntelHex::load( filePath ) )
    {
        // Member durch gelesene Daten initialisieren und pruefen
        if ( m_buffer.size() )
        {
            // Hilfsstrukturzeiger
            m_wlan_config_node = (WLANCONFIGNODE *)&m_buffer[0];

            // 1.
            // WLAN-Konfiguration pruefen

            // Zeigerarray und Wlanstrukturzeiger initialisieren
            std::size_t i = m_wlan_config_node->cnt;
            if ( i > m_wlan_config.size() )
            {
                // Fehler zuweisen
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
            else
            {
                for ( std::size_t j = 0; j < i; ++j )
                {
                    m_wlan_config[j] = (WLANCONFIG *)&m_buffer[m_wlan_config_node->ptrWlanConfigArray + i * sizeof( unsigned int ) + j * sizeof( WLANCONFIG )];
                }

                // Pruefsumme zum Test auf Datenaenderung berechnen.
                m_crc = checksum32bit();
            }
        }
        else
        {
            // Fehler zuweisen
            setErrorState( Error_state::no_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileWlanConfig::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ )
{
    // Pruefsumme zum Test auf Datenaenderung berechnen.
    m_crc = checksum32bit();

    return FileIntelHex::save( filePath, appendToFile, appendEndLine );
}

void FileWlanConfig::DefaultWlanConfig( WLANCONFIG &wc )
{
    std::fill_n( (unsigned char*)&wc, sizeof( WLANCONFIG ), 0 );

    //wc.SSID[33];
    wc.EnableWLAN               = 1;
    wc.Type                     = 0;
    wc.Band                     = 0;
    wc.Channel                  = 0;
    wc.Country                  = 3;
    wc.Security                 = 0;
    wc.Authentication           = 0;
    wc.PairwiseEncryptionMethod = 0;
    wc.GroupEncryptionMethod    = 0;
    //wc.Key[64];
    wc.KeyType                  = 0;
    wc.TxKeyIndex               = 0;
    wc.Autofallback             = 1;
    wc.TxDataRate               = 0;
    wc.RadioPower               = 0;
    wc.PowerLevel               = 2;
    wc.DHCP                     = 1;
    std::string( "0.0.0.0" ).copy( (char*)wc.IpAddress, sizeof( wc.IpAddress ) - 1 );
    wc.Port                     = 8000;
    std::string( "008" ).copy( wc.Hostbits, sizeof( wc.Hostbits ) - 1 );
    std::string( "255.255.255.0" ).copy( wc.SubNetWorkMask, sizeof( wc.SubNetWorkMask ) - 1 );
    std::string( "0.0.0.0" ).copy( wc.GateWay, sizeof( wc.GateWay ) - 1 );
    wc.RemoteAccess             = 1;
    wc.Alive                    = 45;
    //wc.Name[33];
    wc.timeout                  = 60;
    wc.flags                    = 0;
    wc.startupTimeout           = 40;
}

